/*
 * GTK Cervi
 * vim:set sw=4 sta:
 *
 * Copyright (C) 2003  Tomas Janousek <tomi@nomi.cz>
 * See file COPYRIGHT and COPYING
 */

#pragma implementation

#include <cmath>
#include <gdk/gdkkeysyms.h>
#include "game.h"

/*
 * TODO:
 *  Controls :)
 *  Music
 */

namespace std {
    // colors of cerv's
    unsigned int colors[] = {
	0xff0000,
	0x00ff00,
	0x0000ff,
	0xffff00,
	0x00ffff,
	0xff00ff,
	0xff8000,
	0xffffff
    };

    // controls of cerv's
    struct {
	unsigned int left, right;
    } controls[] = {
	{GDK_Left, GDK_Right},
	{GDK_Control_L, GDK_Alt_L},
	{GDK_Tab, GDK_Caps_Lock},
	{0,0},
	{0,0},
	{0,0},
	{0,0},
	{0,0}
    };
    
    // create ncervi cerv's, place them and move them from center
    Game::Game(int ncervi) : _ncollided(0), over(false)
    {
	n_cervi = ncervi;
	for (int i=0; i<n_cervi; i++) {
	    cervi[i] = new Cerv(this,Field::width / 2,Field::height / 2,
		    -M_PI_2 + i*2*M_PI/n_cervi + // position cerv's regularly
			2*M_PI*((rand()%20)-10)/n_cervi/100, // with variation
		    70 + (rand() % 50),10,colors[i],i); // variate speed also
	    cervi[i]->_cspeed = cervi[i]->_speed;
	    cervi[i]->update(75*(n_cervi+2)/2); // keep similar distance
		// between cerv's for any number of cerv's
	    cervi[i]->_cspeed = 1;
	}
    }

    // delete all cervi
    Game::~Game()
    {
	for (int i=0; i<n_cervi; i++) {
	    delete cervi[i];
	}
    }

    // update & draw
    void Game::update(GtkWidget* area, GdkPixmap* pixmap,
	    unsigned long long int ticks)
    {
	GdkGC *gc = gdk_gc_new(((GtkWidget*)area)->window);
	if (!gc) gtk_exit(-1);
	
	if (ticks > 50) // will slow down game, but this can be caused only by
	    ticks = 50; // slow CPU or high load, so in result, user will have
			// control over his Cerv even in these situations
			// (... 20 fps is minimum)

	// update, check collision, draw ...
	for (int i=0; i<n_cervi; i++) {
	    if (cervi[i]->_collision) continue;
	    
	    int _x = (int)nearbyint(cervi[i]->_x),
		_y = (int)nearbyint(cervi[i]->_y);
	    for (unsigned long long int j=0; j<ticks; j++) {
		cervi[i]->update(1);
		int x = (int)nearbyint(cervi[i]->_x),
		    y = (int)nearbyint(cervi[i]->_y);
		if (!(x == _x && y == _y)) {
		    if (!cervi[i]->_collision) cervi[i]->collision();
		    // draw it and set bits in bitfield
		    gdk_rgb_gc_set_foreground(gc,cervi[i]->_color);
		    gdk_draw_point(pixmap,gc,x,y);
		    field._b[Field::pos(x,y)] = 1;
		    gdk_rgb_gc_set_foreground(gc,
			    ((((cervi[i]->_color & 0xff0000)*4/5) & 0xff0000))
			    | ((((cervi[i]->_color & 0xff00)*4/5) & 0xff00)) |
			    ((((cervi[i]->_color & 0xff)*4/5) & 0xff)) );
		    if (!field._b[Field::pos(x-1,y)]) {
			gdk_draw_point(pixmap,gc,x-1,y);
			field._b[Field::pos(x-1,y)] = 1;
		    }
		    if (!field._b[Field::pos(x+1,y)]) {
			gdk_draw_point(pixmap,gc,x+1,y);
			field._b[Field::pos(x+1,y)] = 1;
		    }
		    if (!field._b[Field::pos(x,y-1)]) {
			gdk_draw_point(pixmap,gc,x,y-1);
			field._b[Field::pos(x,y-1)] = 1;
		    }
		    if (!field._b[Field::pos(x,y+1)]) {
			gdk_draw_point(pixmap,gc,x,y+1);
			field._b[Field::pos(x,y+1)] = 1;
		    }
		    
		    // do some "antialiasing"
		    // this is a mess and should be rewritten
		    gdk_rgb_gc_set_foreground(gc,
			    ((((cervi[i]->_color & 0xff0000) >> 1) & 0xff0000))
			    | ((((cervi[i]->_color & 0xff00) >> 1) & 0xff00)) |
			    ((((cervi[i]->_color & 0xff) >> 1) & 0xff)) );
		    if (x == _x && y < _y) {
			gdk_draw_point(pixmap,gc,x+1,y-1);
			gdk_draw_point(pixmap,gc,x-1,y-1);
		    } else if (x == _x && y > _y) {
			gdk_draw_point(pixmap,gc,x+1,y+1);
			gdk_draw_point(pixmap,gc,x-1,y+1);
		    } else if (x < _x && y == _y) {
			gdk_draw_point(pixmap,gc,x-1,y+1);
			gdk_draw_point(pixmap,gc,x-1,y-1);
		    } else if (x > _x && y == _y) {
			gdk_draw_point(pixmap,gc,x+1,y+1);
			gdk_draw_point(pixmap,gc,x+1,y-1);
		    } else if (x > _x && y > _y) {
			gdk_draw_point(pixmap,gc,x+1,y+1);
			gdk_draw_point(pixmap,gc,x-1,y+1);
			gdk_draw_point(pixmap,gc,x+1,y-1);
		    } else if (x < _x && y > _y) {
			gdk_draw_point(pixmap,gc,x-1,y+1);
			gdk_draw_point(pixmap,gc,x+1,y+1);
			gdk_draw_point(pixmap,gc,x-1,y-1);
		    } else if (x > _x && y < _y) {
			gdk_draw_point(pixmap,gc,x+1,y-1);
			gdk_draw_point(pixmap,gc,x+1,y+1);
			gdk_draw_point(pixmap,gc,x-1,y-1);
		    } else if (x < _x && y < _y) {
			gdk_draw_point(pixmap,gc,x-1,y-1);
			gdk_draw_point(pixmap,gc,x-1,y+1);
			gdk_draw_point(pixmap,gc,x+1,y-1);
		    }
		    
		    _x = x;
		    _y = y;
		    GdkRectangle update_rect = {x-1,y-1,3,3};
		    gtk_widget_draw(GTK_WIDGET(area), &update_rect);
		}
		if (cervi[i]->_collision) break;
	    }
	}

	// count collided cerv's
	int ncollision = 0;
	for (int i=0; i<n_cervi; i++) {
	    if (cervi[i]->_collision)
		ncollision++;
	}

	// increase speed if someone collided
	if (ncollision - _ncollided)
	    for (int i=0; i<n_cervi; i++)
		if (!cervi[i]->_collision) {
		    cervi[i]->_speed += (ncollision - _ncollided) * 10;
		}

	// set place numbers
	int maxplace = -1;
	for (int i=0; i<n_cervi; i++) {
	    if (cervi[i]->_collision && cervi[i]->_place > maxplace) {
		maxplace = cervi[i]->_place;
	    }
	}
	if (maxplace == -1) maxplace++;
	maxplace++;
	for (int i=0; i<n_cervi; i++) {
	    if (cervi[i]->_collision && cervi[i]->_place == -1) {
		cervi[i]->_place = maxplace;
	    }
	}

	// is round over ??
	if (ncollision == n_cervi)
	    over = true;

	// if all except one are collided, collide him to end round
	if (ncollision == n_cervi-1 && n_cervi != 1) {
	    for (int i=0; i<n_cervi; i++) {
		if (!cervi[i]->_collision)
		    cervi[i]->_collision = true;
	    }
	}

	_ncollided = ncollision;
	gdk_gc_unref(gc);
    }

    // save place numbers to int pos[n_cervi]
    void Game::result(int *pos)
    {
	int maxplace = -1;
	for (int i=0; i<n_cervi; i++) {
	    if (cervi[i]->_collision && cervi[i]->_place > maxplace) {
		maxplace = cervi[i]->_place;
	    }
	}
	maxplace++;
	for (int i=0; i<n_cervi; i++) {
	    pos[i] = maxplace - cervi[i]->_place;
	}
    }

    // get game speed (average speed of non-collided cerv's)
    double Game::speed()
    {
	double s = 0;
	int n = 0;
	for (int i=0; i<n_cervi; i++) {
	    if (cervi[i]->_collision) continue;
	    s += cervi[i]->_cspeed;
	    n ++;
	}
	if (n) {
	    return s/n;
	}
	return 0;
    }

    // move and/or rotate cerv and check if collided to wall
    void Cerv::update(unsigned long long ticks)
    {
	if (_game->keymap[controls[_n].left])
	    _angle -= ticks * _rotspeed * _cspeed / 150 / 1000;
	if (_game->keymap[controls[_n].right])
	    _angle += ticks * _rotspeed * _cspeed / 150 / 1000;

	while (_angle > 2*M_PI)
	    _angle -= 2*M_PI;
	while (_angle < 0)
	    _angle = 2*M_PI + _angle;
	
	_x += ticks * _cspeed * cos(_angle) / 1000;
	_y += ticks * _cspeed * sin(_angle) / 1000;

	if (_cspeed < _speed) {
	    _cspeed += _speed * ticks / 1000;
	    if (_cspeed > _speed)
		_cspeed = _speed;
	}

	if (_x < 1) {
	    _x = 1;
	    _collision = true;
	}
	if (_y < 1) {
	    _y = 1;
	    _collision = true;
	}
	if (_x > (Field::width-2)) {
	    _x = Field::width-2;
	    _collision = true;
	}
	if (_y > (Field::height-2)) {
	    _y = Field::height-2;
	    _collision = true;
	}
    }

    // check if collided to himself or another cerv
    int Cerv::collision()
    {
	if (!_collision) {
	    if (_angle >= 0 && _angle < M_PI_2) {
		if    (_game->field._b[Field::pos((int)nearbyint(_x)+2,
			    (int)nearbyint(_y))]
		    || _game->field._b[Field::pos((int)nearbyint(_x)+1, //  |
			    (int)nearbyint(_y)+1)]			// -+-
		    || _game->field._b[Field::pos((int)nearbyint(_x),	//  |/
			    (int)nearbyint(_y)+2)]) {
		    _collision = true;
		}
	    } else if (_angle >= M_PI_2 && _angle < M_PI) {
		if    (_game->field._b[Field::pos((int)nearbyint(_x)-2,
			    (int)nearbyint(_y))]
		    || _game->field._b[Field::pos((int)nearbyint(_x)-1, //  |
			    (int)nearbyint(_y)+1)]			// -+-
		    || _game->field._b[Field::pos((int)nearbyint(_x),	// \|
			    (int)nearbyint(_y)+2)]) {
		    _collision = true;
		}
	    } else if (_angle >= M_PI && _angle < M_PI+M_PI_2) {
		if    (_game->field._b[Field::pos((int)nearbyint(_x)-2,
			    (int)nearbyint(_y))]
		    || _game->field._b[Field::pos((int)nearbyint(_x)-1, // /|
			    (int)nearbyint(_y)-1)]			// -+-
		    || _game->field._b[Field::pos((int)nearbyint(_x),	//  |
			    (int)nearbyint(_y)-2)]) {
		    _collision = true;
		}
	    } else if (_angle >= M_PI+M_PI_2 && _angle < 2*M_PI) {
		if    (_game->field._b[Field::pos((int)nearbyint(_x)+2,
			    (int)nearbyint(_y))]
		    || _game->field._b[Field::pos((int)nearbyint(_x)+1, //  |\*
			    (int)nearbyint(_y)-1)]			// -+-
		    || _game->field._b[Field::pos((int)nearbyint(_x),	//  |
			    (int)nearbyint(_y)-2)]) {
		    _collision = true;
		    // * - \ would make comment continue on next line, so *
		}
#ifdef DEBUG
	    } else {
		// this cannot happen if EVERYTHING is right
		cerr << "This cannot happen!!! " << __FILE__ << " - " <<
		    __LINE__ << endl;
#endif /* DEBUG */
	    }
	}
	return _collision;
    }
}
